/*Listing 13-5.  C Portion of Package, SERIAL.C*/
/*---------------------------------------------------------------------*/
/*
 * Filename:        S E R I A L . C
 * Purpose:         An asynchronous communication package for
 *                  IBM PC and compatibles.  Supports asynchronous
 *                  communication adapters based on Intel 8250
 *                  UART (or compatible).
 * Language:        Microsoft C 4.0
 * Usage:           This is a set of routines that provides basic
 *                  functions for asynchronous communication.
 *                  Call s_setvals to set parameters and s_getvals
 *                  see parameters values.  Call s_setup to setup
 *                  port for communications.  Use s_sendchar and
 *                  s_rcvchar to send and receive characters
 *                  respectively.
 *
 * Copyright  1986, 1987 Nabajyoti Barkakati
 *                         Silver Sprint, MD 20904
 * Right to use, copy, and modify this code is granted for personal
 * noncommercial use, provided that this copyright disclosure is   
 * retained on all copies.  All other rights reserved.
 *
 * Data Started:        10-DEC-1986
 * Revisions:  V1.00     9-FEB-1987--First working version. (NB)
*/
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include "idecls.h"
#include "serial.h"

/* Function to turn on interrupt whose "Interrupt Enable Number:
 * is 'i', in case it has been disabled.  For example, the 
 * THRE interrupt is disable when an XOFF is received from the 
 * remote system.
 */

#define turnon_int(i,j)  if(((j=inp(IER))&i)==0)outp(IER,(j|i))

#define report_error(s)  fprintf(stderr,s)

/*Global status indicators */

int  s_linestatus, s_modemstatus, s_rbocount, s_totalcount,
     s_rlscount, s_rdacount, s_trmtycount, s_mscount; 

/* Old vector pointers for COM1 and COM2 */
	
void (_interrupt _far *oldcom1_vect)();
void (_interrupt _far *oldcom2_vect)();

/* Local function prototypes */

static void s_rls ( void );
static void s_rda ( void );
static void s_trmty ( void );
static void s_ms ( void );
static void s_intinit ( void );
static void s_intoff ( void );


static QTYPE *txq = (QTYPE *) NULL, *rxq = (QTYPE *) NULL;
static union REGS xr, yr;
static struct SREGS sr;

int comport = 0;

static int   port_number = 0,
             baudrate = 7,
             parity = 0,
             stopbits = 1,
             wordlength = 3,
             txqsize = 1024,
             rxqsize = 2048,
             bit7_1_rx = FALSE,
             bit7_0_rx = TRUE,
             bit7_1_tx = FALSE,
             bit7_0_tx = FALSE,
             pckd_comparams = 0x83,
             int_enable_mask = 0xef,
             int_disable_mask = 0x10;


/*-------------------------------------------------------------*/
/* s _ m a i n d l r
 * Main interrupt handler for all serial port interrupts.
 */

void interrupt cdecl far s_mainhndlr ( void )
{
    register int int_id, intmask;
    while (TRUE) 
    {
	/* Read the interrupt identification register, IIR */

 	int_id = inp(IIR);
 	if (bit0(int_id) == 1) 
   	{
		/* If bit 0 is 1, then no interrupts pending.  Send an end of
		 * interrupt signal to the 8259A Programmable Interrupt
		 * Controller and then return.
		 */
		 
             	outp(P8259_0, END_OF_INT);
             	return;
	}
	
	s_totalcount++;
	if (int_id >= RXDATAREADY) 
		turnon_int( THREINT, intmask );

	/* Process interrupt according to id.
	 * The following list is in increasing order of priority.
	 */
	 
         switch (int_id) 
         {
      		case MDMSTATUS:		s_ms();
                                 	break;
                case TXREGEMPTY: 	s_trmty();
                                 	break;
		case RXDATAREADY:  	s_rda();
                                 	break;
		case RLINESTATUS:   	s_rls();
                                 	break;
                                 
		/* Just fall through if id is none of the above */
        }
    }
}

/*------------------------------------------------------------*/
/* s_ r l s
 * Process a "receive line status" interrupt
 */

static void s_rls ( void )
{
	/* Read line status and save it in 's_linestatus' */
	
	s_linestatus = inp(LSR);
	s_rlscount++;
}

/*-----------------------------------------------------------*/
/* s _ r d a 
 * Process a "receive data available" interrupt
 */

static void s_rda ( void )
{
char c;
    
 	/* Read from comport */
 	
 	c = (char) inp ( comport );
	s_rdacount++;
		
	if(q_puton(rxq, &c) == 0)
		/* Increment receive buffer overflow count */
		s_rbocount++;

}

/*---------------------------------------------------------------*/
/* s _ t r m t y 
 * Process a "transmit holding register empty" interrupt
 */

static void s_trmty ( void )
{
    char c;
    register int ierval;

    s_trmtycount++;

    /* Put a character into the transmit holding register */

    if( q_getfrom(txq, &c) != NULL){
        if(bit7_1_tx) c |= 0x80;
        if(bit7_0_tx) c &= 0x7f;
        outp(comport, c);
        return;
    }
    
    /* Nothing to send--turn off THRE interrupts */
    
    ierval = inp(IER);
    if (ierval & THREINT) outp(IER, ierval & THREOFF);
}

/*-----------------------------------------------------------------*/
/* s _ m s
 * Process a "modem status" interrupt
 */

static void s_ms ( void )
{
    /* Read moden status and save in 's_modemstatus' */

    s_modemstatus = inp(MSR);
    s_mscount++;
}

/*---------------------------------------------------------------*/
/* s _ c o m _  i n i t
 *
 *
 */
 
void s_com_init( void )
{
   oldcom1_vect = _dos_getvect( 12 );
   oldcom2_vect = _dos_getvect( 11 );
   _dos_setvect( 12, s_mainhndlr );
   _dos_setvect( 11, s_mainhndlr );
   
   s_setcommparams();
   
   if (  ( txq = q_setup( txqsize, sizeof( char ) ) ) == NULL )
   {
      report_error("Error creating transmit queue!\n");
      return;
   }
   if (  ( rxq = q_setup( rxqsize, sizeof( char ) ) ) == NULL )
   {
      report_error("Error creating receive queue!\n");
      return;
   }
      
   /* Reset all counts and flags. */

    s_rbocount = 0;
    s_totalcount = 0;
    s_rlscount = 0;
    s_rdacount = 0;
    s_trmtycount = 0;
    s_mscount = 0;
    
/*    outp( *(BIOS_DATA + _COM1 + 1), 0x0F );
    outp( *(BIOS_DATA + _COM2 + 1), 0x0F );*/
}
   
  
#ifdef SKIP
/*-------------------------------------------------------------------*/
/* s _ i n t i n i t
 * Start up interrupts from the serial board.  Also set up
 * 8259A to accept the interrupts.  This routine should be
 * invoked with the interrupts turned off, i.e., inside a
 * (CLI---STI) pair.
 */

static void s_intinit ( void ) 
{
int intmask;

    /* Set up moden control register (port = MCR) */

    outp(MCR, MCRALL);

    /* Enable all interrupts on the serial card (port = IER) */

    outp(IER, IERALL);

    /* Read 8259A's interrupt mask register and write it back after
     * ANDing with int_enable_mask.
     */
     
    intmask = inp(P8259_1) & int_enable_mask;
    outp(P8259_1, intmask);
}
#endif

/*------------------------------------------------------------------*/
/* s _ c l e a n u p
 * Cleanup after comm session is done.  Turns off all interrupts.
 */

void s_cleanup ( void )
{
	/* Turn off interrupts from serial card */

	_disable();
	s_intoff();
	_enable();

	/*  Restore original interrupt vectors 
	 *
	 *  Note - int_numbers 12 and 11 correspond to old_vect_ptr_1 and 2
	 *  respectively.
	 */

	_dos_setvect( 12, oldcom1_vect );
	_dos_setvect( 11, oldcom2_vect );

	/* Deallocate space used by the transmit and receive queues */

 	q_reset( txq );
	q_reset( rxq );
}

/*----------------------------------------------------------------*/
/* s _ i n t o f f
 * Turn off all interrupts after the comm session is done.
 * Should be called with interrupts cleared.
 */

static void s_intoff ( void )
{
int intmask = 0x18;
  
	/* First reset the Interrupt Enable Register on the comm card */

	outp( *(BIOS_DATA + _COM1 + 1), IEROFF);
	outp( *(BIOS_DATA + _COM2 + 1), IEROFF);

	/* Turn off all bits of Modem Control Register */

	outp( *(BIOS_DATA + _COM1 + 4), MCROFF);
	outp( *(BIOS_DATA + _COM2 + 4), MCROFF);

	/* Next disable 8259A from recognizing ints at this IRQ level  */

	intmask = inp(P8259_1) | intmask;
	outp(P8259_1, intmask);
}

/*-------------------------------------------------------------*/
/* s _ s e t c o m m p a r a m s
 * Set up basic communication parameters by using BIOS interrupt    
 * number 14h, function 0 (ah=0).
 */

void s_setcommparams ( void )
{
	/* Set up communication port parameters.  Use BIOS INT 14h, AH=0 */

	pckd_comparams = (baudrate << 5) | (parity << 3) |
                  (stopbits << 2) | (wordlength);
	xr.h.ah = 0;
	xr.h.al = (unsigned char) pckd_comparams;
	xr.x.dx = _COM1;
	int86(BIOS_RS232, &xr, &yr);
	xr.x.dx = _COM2;
	int86(BIOS_RS232, &xr, &yr);
}

/*---------------------------------------------------------*/
/* s _ s e n d c h a r
 * Puts a character into transmit queue.  Returns 1 if all's ok,
 * o if there were problems.
 */

#pragma check_stack(off)

int s_sendchar( int ch )
{
int retval, intmask;

	retval = q_puton(txq, (char *) &ch );

	/* Turn on THRE interrupt if it's off */
	
	turnon_int(THREINT,intmask);
	return(retval);
}

#pragma check_stack(on)

/*---------------------------------------------------------*/                  
/* s _ r c v c h a r 
 * Returns a character from the receive queue.
 * Returns -1 of queue is empty.
 */

int s_rcvchar ( void )
{
int ch;
	if ( q_getfrom(rxq, (char *) &ch) == NULL)
		return(-1);
	else 
		return(ch);
}

/*---------------------------------------------------------*/
/* s _ t x q e m p t y
 * Returns 1 if txq (transmit queue) is empty, else returns 0.
 * Need when sending file.
 */

int s_txqempty ( void )
{
    if (txq->count == 0)
       return (TRUE);
    else
         return (FALSE);
}

/*------------------END OF SERIAL.C----------------------------------------*/
